from abc import ABC, abstractmethod
from typing import Optional, List, Dict

from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import GraphNode
from sampo.schemas.resources import Worker
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator


class Timeline(ABC):

    @abstractmethod
    def schedule(self,
                 task_index: int,
                 node: GraphNode,
                 node2swork: Dict[GraphNode, ScheduledWork],
                 passed_agents: List[Worker],
                 contractor: Contractor,
                 assigned_start_time: Optional[Time],
                 assigned_time: Optional[Time],
                 work_estimator: Optional[WorkTimeEstimator] = None) -> Time:
        ...

    @abstractmethod
    def find_min_start_time(self,
                            node: GraphNode,
                            worker_team: List[Worker],
                            id2swork: Dict[str, ScheduledWork],
                            parent_time: Time = Time(0),
                            work_estimator: Optional[WorkTimeEstimator] = None) -> Time:
        ...

    @abstractmethod
    def update_timeline(self,
                        task_index: int,
                        finish_time: Time,
                        node: GraphNode,
                        node2swork: Dict[GraphNode, ScheduledWork],
                        worker_team: List[Worker]):
        ...

    @abstractmethod
    def __getitem__(self, item):
        ...



from enum import Flag, auto
from typing import Optional, Union

from matplotlib.pyplot import Figure as MatplotlibFig
from plotly.graph_objects import Figure as PlotlyFig


class VisualizationMode(Flag):
    NoFig = 0
    ShowFig = auto()
    SaveFig = auto()
    ReturnFig = auto()


def visualize(fig: Union[PlotlyFig, MatplotlibFig], mode: VisualizationMode, file_name: str = None) \
        -> Optional[Union[PlotlyFig, MatplotlibFig]]:
    """
    Visualizes the figure according to the provided settings
    :param fig: The figure
    :param mode: Visualisation mode. Can be bitwise-or (|) of several modes.
    :param file_name: Optional name of a saved file. Passed, if SaveFig in mode.
    :return: The figure, if ReturnFig in mode. Otherwise, None.
    """
    if VisualizationMode.SaveFig in mode:
        if isinstance(fig, PlotlyFig):
            fig.write_image(file_name)
        else:
            fig.savefig(file_name)
    if VisualizationMode.ShowFig in mode:
        fig.show()
    if VisualizationMode.ReturnFig in mode:
        return fig
    return None



from typing import List, Tuple

import numpy as np

from sampo.scheduler.base import Scheduler
from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.timeline.just_in_time_timeline import JustInTimeTimeline
from sampo.schemas.contractor import WorkerContractorPool, Contractor
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import ScheduledWork, Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator

ChromosomeType = Tuple[np.ndarray, np.ndarray, np.ndarray]


def convert_schedule_to_chromosome(wg: WorkGraph,
                                   work_id2index: dict[str, int], worker_name2index: dict[str, int],
                                   contractor2index: dict[str, int], contractor_borders: np.ndarray,
                                   schedule: Schedule, order: list[GraphNode] | None = None) -> ChromosomeType:
    """
    Receive result of scheduling algorithm and transform it to chromosome

    :param wg:
    :param work_id2index:
    :param worker_name2index:
    :param contractor2index:
    :param contractor_borders:
    :param schedule:
    :param order: if passed, specifies the node order that should appear in the chromosome
    :return:
    """

    order: list[GraphNode] = order if order is not None else [work for work in schedule.works
                                                              if not wg[work.work_unit.id].is_inseparable_son()]

    # order works part of chromosome
    order_chromosome: np.ndarray = np.array([work_id2index[work.work_unit.id] for work in order])

    # convert to convenient form
    schedule = schedule.to_schedule_work_dict

    # resources for works part of chromosome
    # +1 stores contractors line
    resource_chromosome = np.zeros((len(worker_name2index) + 1, len(order_chromosome)), dtype=int)

    for node in order:
        node_id = node.work_unit.id
        index = work_id2index[node_id]
        for resource in schedule[node_id].workers:
            res_count = resource.count
            res_index = worker_name2index[resource.name]
            res_contractor = resource.contractor_id
            resource_chromosome[res_index, index] = res_count
            resource_chromosome[-1, index] = contractor2index[res_contractor]

    resource_border_chromosome = np.copy(contractor_borders)

    return order_chromosome, resource_chromosome, resource_border_chromosome


def convert_chromosome_to_schedule(chromosome: ChromosomeType, worker_pool: WorkerContractorPool,
                                   index2node: dict[int, GraphNode],
                                   index2contractor: dict[int, Contractor],
                                   worker_pool_indices: dict[int, dict[int, Worker]],
                                   spec: ScheduleSpec,
                                   timeline: Timeline | None = None,
                                   assigned_parent_time: Time = Time(0),
                                   work_estimator: WorkTimeEstimator = None,) \
        -> tuple[dict[GraphNode, ScheduledWork], Time, Timeline]:
    node2swork: dict[GraphNode, ScheduledWork] = {}

    if not isinstance(timeline, JustInTimeTimeline):
        timeline = JustInTimeTimeline(worker_pool)
    works_order = chromosome[0]
    works_resources = chromosome[1]

    for order_index, work_index in enumerate(works_order):
        node = index2node[work_index]
        if node.id in node2swork and not node.is_inseparable_son():
            continue

        work_spec = spec.get_work_spec(node.id)

        resources = works_resources[:-1, work_index]
        contractor_index = works_resources[-1, work_index]
        contractor = index2contractor[contractor_index]
        worker_team: List[Worker] = [worker_pool_indices[worker_index][contractor_index]
                                     .copy().with_count(worker_count)
                                     for worker_index, worker_count in enumerate(resources)
                                     if worker_count > 0]

        # apply worker spec
        Scheduler.optimize_resources_using_spec(node.work_unit, worker_team, work_spec)

        st = timeline.find_min_start_time(node, worker_team, node2swork, assigned_parent_time, work_estimator)

        if order_index == 0:  # we are scheduling the work `start of the project`
            st = assigned_parent_time  # this work should always have st = 0, so we just re-assign it

        # finish using time spec
        finish_time = timeline.schedule(order_index, node, node2swork, worker_team, contractor,
                                        st, work_spec.assigned_time, assigned_parent_time, work_estimator)

        timeline.update_timeline(order_index, finish_time, node, node2swork, worker_team)

    schedule_start_time = min([swork.start_time for swork in node2swork.values() if
                               len(swork.work_unit.worker_reqs) != 0])

    return node2swork, schedule_start_time, timeline



from copy import deepcopy
from itertools import chain
from random import Random
from typing import Callable

from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.requirements import WorkerReq


def _extend_str_fields(new_uniq_count: int, wg: WorkGraph, rand: Random,
                       uniq_finder_str: Callable[[WorkGraph], list[str]],
                       update_node: Callable[[GraphNode, dict[str, list[str]], Random], None]) -> WorkGraph:
    uniq_str = uniq_finder_str(wg)
    if new_uniq_count <= len(uniq_str):
        return wg
    wg = deepcopy(wg)
    rand.shuffle(uniq_str)
    names_plus_one = set(uniq_str[:new_uniq_count % len(uniq_str)])
    suffixes = [' #' + str(i + 1) for i in range(new_uniq_count // len(uniq_str))]
    name_to_suffixes = {name: (suffixes + (['_0'] if name in names_plus_one else [])) for name in uniq_str}
    [update_node(node, name_to_suffixes, rand) for node in wg.nodes]
    return wg


def _get_uniq_resource_kinds(wg: WorkGraph) -> list[str]:
    uniq_res_kinds = list(set(req.kind for req in chain(*[n.work_unit.worker_reqs for n in wg.nodes])))
    return uniq_res_kinds


def _update_resource_names(node: GraphNode, name_to_suffixes: dict[str, list[str]], rand: Random):
    new_req: list[WorkerReq] = []
    for res_req in node.work_unit.worker_reqs:
        kind = res_req.kind
        options = name_to_suffixes[kind]
        kind = kind + options[rand.randint(0, len(options)-1)]
        new_req.append(WorkerReq(kind, res_req.volume, res_req.min_count, res_req.max_count, res_req.name))
    node.work_unit.worker_reqs = new_req


def _get_uniq_work_names(wg: WorkGraph) -> list[str]:
    start_finish_names = {wg.start.work_unit.name, wg.finish.work_unit.name}
    uniq_work_names = list(set(node.work_unit.name for node in wg.nodes) - start_finish_names)
    return uniq_work_names


def _update_work_name(node: GraphNode, name_to_suffixes: dict[str, list[str]], rand: Random) -> None:
    name = node.work_unit.name
    if name in name_to_suffixes:
        options = name_to_suffixes[name]
        node.work_unit.name = name + options[rand.randint(0, len(options)-1)]


def extend_resources(uniq_resources: int, wg: WorkGraph, rand: Random) -> WorkGraph:
    """
    Increases the number of unique resources in WorkGraph
    :param uniq_resources: the amount to which you need to increase
    :param wg: original WorkGraph
    :param rand: Number generator with a fixed seed, or None for no fixed seed
    :return: modified WorkGraph
    """
    return _extend_str_fields(uniq_resources, wg, rand, _get_uniq_resource_kinds, _update_resource_names)


def extend_names(uniq_activities: int, wg: WorkGraph, rand: Random) -> WorkGraph:
    """
     Increases the number of unique work names in WorkGraph
    :param uniq_activities:  the amount to which you need to increase
    :param wg: original WorkGraph
    :param rand: Number generator with a fixed seed, or None for no fixed seed
    :return: modified WorkGraph
    """
    return _extend_str_fields(uniq_activities, wg, rand, _get_uniq_work_names, _update_work_name)



from dataclasses import dataclass
from random import Random
from typing import Optional

import numpy as np

from sampo.schemas.identifiable import Identifiable
from sampo.schemas.interval import IntervalGaussian
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.types import AgentId


@dataclass
class Resource(AutoJSONSerializable['Equipment'], Identifiable):
    """
    A class summarizing the different resources used in the work: Human resources, equipment, materials, etc.
    """
    pass


# TODO describe "productivity"
@dataclass
class Worker(Resource):
    """
    A class dedicated to human resources
    :param count: the number of people in this resource
    :param contractor_id: Contractor id if resources are added directly to the contractor
    :param productivity: Contractor id if resources are added directly to the contractor
    """

    def __init__(self,
                 id: str,
                 name: str,
                 count: int,
                 contractor_id: Optional[str] = "",
                 productivity: Optional[IntervalGaussian] = IntervalGaussian(1, 0, 1, 1),
                 cost_one_unit: Optional[float] = None):
        super(Worker, self).__init__(id, name)
        self.count = count
        self.contractor_id = contractor_id
        self.productivity = productivity if productivity is not None else IntervalGaussian(1, 0, 1, 1)
        self.cost_one_unit = cost_one_unit if cost_one_unit is not None else self.productivity.mean * 10

    ignored_fields = ['productivity']

    # TODO: describe the function (description, return type)
    def copy(self):
        return Worker(id=self.id,
                      name=self.name,
                      count=self.count,
                      contractor_id=self.contractor_id,
                      productivity=self.productivity)

    def with_count(self, count: int) -> 'Worker':
        self.count = count
        return self

    def get_cost(self) -> float:
        """Returns cost of this worker entry"""
        return self.cost_one_unit * self.count

    # TODO: describe the function (description, return type)
    def get_agent_id(self) -> AgentId:
        return self.contractor_id, self.name

    # TODO: describe the function (description, return type)
    def get_static_productivity(self) -> float:
        return self.productivity.mean * self.count

    # TODO: describe the function (description, return type)
    def get_stochastic_productivity(self, rand: Optional[Random] = None) -> float:
        return self.productivity.rand_float(rand) * self.count

    # TODO: describe the function (description, return type)
    def __repr__(self):
        return f'{self.count} {self.name}'

# TODO: describe the class (description)
@dataclass
class ConstructionObject(Resource):
    pass

# TODO: describe the class (description, parameters)
@dataclass(init=False)
class EmptySpaceConstructionObject(ConstructionObject):
    id: str = "00000000000000000"
    name: str = "empty space construction object"

# TODO: describe the class (description)
@dataclass
class Equipment(Resource):
    pass

# TODO: describe the class (description)
@dataclass
class Material(Resource):
    pass



from collections import defaultdict
from copy import copy
from typing import List, Dict, Optional, Union

from sampo.schemas.resources import Worker
from sampo.schemas.time import Time
from sampo.schemas.types import WorkerName
from sampo.schemas.works import WorkUnit


class WorkSpec:
    """
    Here are the container for externally given terms, that the resulting `ScheduledWork` should satisfy.
    Must be used in schedulers.
    :param chain: the chain of works, that should be scheduled one after another, e.g. inseparable,
    that starts from this work. Now unsupported.
    :param assigned_workers: predefined worker team (scheduler should assign this worker team to this work)
    :param assigned_time: predefined work time (scheduler should schedule this work with this execution time)
    """
    chain: Optional[List[WorkUnit]] = None  # TODO Add support
    assigned_workers: Dict[WorkerName, int] = {}
    assigned_time: Optional[Time] = None


class ScheduleSpec:
    """
    Here are the container for externally given terms, that Schedule should satisfy.
    Must be used in schedulers.
    :param work2spec: work specs
    """
    _work2spec: Dict[str, WorkSpec] = defaultdict(WorkSpec)

    def set_exec_time(self, work: Union[str, WorkUnit], time: Time) -> 'ScheduleSpec':
        if isinstance(work, WorkUnit):
            work = work.id
        self._work2spec[work].assigned_time = time
        return self

    def assign_workers_dict(self, work: str, workers: Dict[WorkerName, int]) -> 'ScheduleSpec':
        if isinstance(work, WorkUnit):
            work = work.id
        self._work2spec[work].assigned_workers = copy(workers)
        return self

    def assign_workers(self, work: str, workers: List[Worker]) -> 'ScheduleSpec':
        if isinstance(work, WorkUnit):
            work = work.id
        self._work2spec[work].assigned_workers = {worker.name: worker.count for worker in workers}
        return self

    def get_work_spec(self, work_id: str) -> WorkSpec:
        return self._work2spec[work_id]

